/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: phg-to-p4est.c,v 1.75 2021/06/23 02:33:14 zlb Exp $ */

#include <p8est_vtk.h>
#include <p8est_iterate.h>
#include <p8est_mesh.h>
#include <p8est_extended.h>
#include <p8est_bits.h>

#define NEED_GetVariableArgs
#include <phg.h>
#include "phg-to-p4est.h"

static void
xyz_to_ref(GRID *g, INT eno, const FLOAT x[], FLOAT xref[])
/* returns in xref[] the coordinates of point x[] in the ref. element (-1,1)^3.
 * dx[] is an optional array which, if not NULL, returns the relative size
 * of the element w.r.t. the reference element. */
{
    FLOAT (*cuboid)[3] = g->elems[eno]->cuboid;

    xref[0] = (x[0] - cuboid[0][0]) / (-cuboid[0][0] + cuboid[1][0]);
    xref[1] = (x[1] - cuboid[0][1]) / (-cuboid[0][1] + cuboid[1][1]);
    xref[2] = (x[2] - cuboid[0][2]) / (-cuboid[0][2] + cuboid[1][2]);
}

static FLOAT *
ref_to_xyz(GRID *g, INT eno, FLOAT xref[], FLOAT x[])
{
    FLOAT (*cub)[Dim] = phgGeomGetCuboid(g, g->elems[eno], NULL);

    x[0] = cub[0][0] + xref[0] * (cub[1][0] - cub[0][0]);
    x[1] = cub[0][1] + xref[1] * (cub[1][1] - cub[0][1]);
    x[2] = cub[0][2] + xref[2] * (cub[1][2] - cub[0][2]);

    return x;
}

static FLOAT
basis1D(int n, int k, FLOAT x, const FLOAT *points, FLOAT *coeff)
/* returns value of the k-th 1D basis function at the point x */
{
    int i;
    FLOAT d, pk;

#if 0	/* not necessary, it's done in DOF_TYPE_init() */
#if USE_OMP
# pragma omp critical
#endif	/* USE_OMP */
#endif
    if (*coeff == 0.) {
	d = 1.0;
	pk = points[k];
	for (i = 0; i < k; i++)
	    d *= (pk - points[i]);
	for (i = k + 1; i <= n; i++)
	    d *= (pk - points[i]);
	*coeff = 1.0 / d;
    }

    d = *coeff;
    for (i = 0; i < k; i++)
	d *= (x - points[i]);
    for (i = k + 1; i <= n; i++)
	d *= (x - points[i]);

    return d;
}

static FLOAT
grad1D(int n, int k, FLOAT x, const FLOAT *points, FLOAT coeff, FLOAT bas)
/* returns value of derivative of the k-th 1D basis function at the point x */
{
    int i, j;
    FLOAT d;

    for (j = 0; j <= n; j++) {
	if (j == k)
	    continue;
	if (x == points[j])
	    break;
    }
    if (j <= n) {
	/* x == points[j] */
	d = coeff;
	for (i = 0; i <= n; i++) {
	    if (i == j || i == k)
		continue;
	    d *= (x - points[i]);
	}
    }
    else {
	/* x != points[i], forall i */
	d = 0.0;
	for (i = 0; i <= n; i++) {
	    if (i == k)
		continue;
	    d += bas / (x - points[i]);
	}
    }

    return d;
}

static FLOAT inline
power(FLOAT x, int p)
/* returns x^p */
{
    if (p < 0)
	return 0.0;
    else if (p == 0)
	return 1.0;
    /* FIXME: faster code! */
    return Pow(x, (FLOAT)p);
}

static void
basisFunc(DOF *u, FLOAT x, FLOAT y, FLOAT z, int index, FLOAT *funcValue)
/* 标准单元上的基函数 */
{
    if (u->type->points != NULL) {
#define TENSOR_PRODUCT \
	int i, j, k, ox, oy, oz; \
	FLOAT *px, *py, *pz, *cx, *cy, *cz; \
	if (u->type->orders == NULL) \
	    phgError(1, "%s:%d: unimplemented.\n", __FILE__, __LINE__); \
	ox = u->type->orders[0]; \
	oy = u->type->orders[1]; \
	oz = u->type->orders[2]; \
	/* Note: index := i + (j + k * (oy+1)) * (ox+1) */ \
	i = index % (ox + 1); \
	index /= (ox + 1); \
	j = index % (oy + 1); \
	k = index / (oy + 1); \
	px = u->type->points; \
	py = px + ox + 1; \
	pz = py + oy + 1; \
	cx = u->type->work; \
	cy = cx + ox + 1; \
	cz = cy + oy + 1;
	TENSOR_PRODUCT
	*funcValue = basis1D(ox, i, x, px, cx + i) *
		     basis1D(oy, j, y, py, cy + j) *
		     basis1D(oz, k, z, pz, cz + k);
    }
    else {
	/* Note: the (i,j,k) is encoded in work[index] as:
		i + (j + k * (n + 1)) * (n + 1) */
	int i, j, k, a, n = u->type->order;
	a = (int)(u->type->work[index] + 0.5);
	i = a % (n + 1);
	a /= n + 1;
	j = a % (n + 1);
	k = a / (n + 1);
	*funcValue = power(x, i) * power(y, j) * power(z, k);
    }
}

static void
gradBasisFunc(DOF *u, FLOAT x, FLOAT y, FLOAT z, int index, FLOAT *grad)
/* 标准单元基函数梯度 */
{
    if (u->type->points != NULL) {
	FLOAT bx, by, bz;
	TENSOR_PRODUCT
#undef TENSOR_PRODUCT
	bx = basis1D(ox, i, x, px, cx + i);
	by = basis1D(oy, j, y, py, cy + j);
	bz = basis1D(oz, k, z, pz, cz + k);
	grad[0] = grad1D(ox, i, x, px, cx[i], bx) * by * bz;
	grad[1] = grad1D(oy, j, y, py, cy[j], by) * bx * bz;
	grad[2] = grad1D(oz, k, z, pz, cz[k], bz) * bx * by;
    }
    else {
	/* Note: the (i,j,k) is encoded in work[index] as:
		i + (j + k * (n + 1)) * (n + 1) */
	int i, j, k, a, n = u->type->order;
	a = (int)(u->type->work[index] + 0.5);
	i = a % (n + 1);
	a /= n + 1;
	j = a % (n + 1);
	k = a / (n + 1);
	grad[0] = i == 0 ? 0. : i * power(x, i - 1) * power(y, j) * power(z, k);
	grad[1] = j == 0 ? 0. : j * power(x, i) * power(y, j - 1) * power(z, k);
	grad[2] = k == 0 ? 0. : k * power(x, i) * power(y, j) * power(z, k - 1);
    }
}

/*------------------------ The QC interface for P4EST ------------------------*/

static void
to_ref(QCACHE *qc, INT eno, FLOAT x[], FLOAT xref[])
{
    DOF *u = qc->fe;
    GRID *g = u->g;

    xyz_to_ref(g, eno, x, xref);
}

static void
from_ref(QCACHE *qc, INT eno, FLOAT xref[], FLOAT x[])
{
    DOF *u = qc->fe;
    GRID *g = u->g;

    ref_to_xyz(g, eno, xref, x);
}


static int
b_dim(QCACHE *qc) { return ((DOF *)qc->fe)->type->dim; }

static int
f_dim(QCACHE *qc, void *fe) { return ((DOF *)fe)->dim; }

static int
n_bas(QCACHE *qc, INT eno)
{
    return ((DOF *)qc->fe)->type->nbas;
}

static void
bas(QCACHE *qc, INT eno, const FLOAT xref[], FLOAT *buffer, int stride)
{
    int j, nb = n_bas(qc, eno);

    for (j = 0; j < nb; j++, buffer += stride)
	basisFunc(qc->fe, xref[0], xref[1], xref[2], j, buffer);
}

static void
grd(QCACHE *qc, INT eno, const FLOAT xref[], FLOAT *buffer, int stride)
{
    int j, k, dim = b_dim(qc), nb = n_bas(qc, eno);
    FLOAT dx, dy, dz;
    FLOAT (*cuboid)[3] = ((DOF *)(qc->fe))->g->elems[eno]->cuboid;

    dx = 1.0 / (-cuboid[0][0] + cuboid[1][0]);
    dy = 1.0 / (-cuboid[0][1] + cuboid[1][1]);
    dz = 1.0 / (-cuboid[0][2] + cuboid[1][2]);
    for (j = 0; j < nb; j++, buffer += stride) {
	for (k = 0; k < dim; k++) {
	    gradBasisFunc(qc->fe, xref[0], xref[1], xref[2], j, buffer + k*Dim);
	    buffer[k * Dim + 0] *= dx;
	    buffer[k * Dim + 1] *= dy;
	    buffer[k * Dim + 2] *= dz;
	}
    }
    return;
}

static void
fun(QCACHE *qc, void *fe, INT eno, const FLOAT x[], FLOAT *buffer)
{
    GRID *g = ((DOF *)fe)->g;
    ELEMENT *e = g->elems[eno];

    phgDofEval(fe, e, x, buffer);
}

QDESC QD_P4EST_ = {"P4EST", to_ref, from_ref, b_dim, f_dim, n_bas, bas, grd,
			fun, Dim, Dim};

/*-------------------- End of the QC interface for P4EST --------------------*/

/*--------------------------- PHG_to_p8est wrapper ---------------------------*/

INT p8est_i;	/* loop variable used by ForAllElements */

/* DOF types */

static DOF_TYPE DOF_ANALYTIC_ = {
	"ANALYTIC",	/* name */
	NULL,		/* grad_type */
	NULL,		/* points */
	NULL,		/* work */
	NULL,		/* orders */
	1,		/* dim */
	0,		/* nbas */
	-1,		/* order */
	FE_L2		/* FE_SPACE */ };

DOF_TYPE **DOF_DGQn = NULL, **DOF_DGPn = NULL, *DOF_ANALYTIC = &DOF_ANALYTIC_;
DOF_TYPE *DOF_DEFAULT = NULL;

static void
DOF_TYPE_init(void)
{
    int i, j, k, l, o, n = DOF_TYPE_ORDER_MAX + 1;
    DOF_TYPE *type;
    char name[128];
    FLOAT d, *p;

    DOF_DGQn = phgAlloc(n * sizeof(*DOF_DGQn));
    DOF_DGPn = phgAlloc(n * sizeof(*DOF_DGPn));
    for (o = 0; o < n; o++) {
	/* DGQn */
	DOF_DGQn[o] = type = phgCalloc(1, sizeof(*type));
	sprintf(name, "DGQ%d", o);
	type->name = strdup(name);	/* name */
	type->grad_type = type;		/* grad_type */
	/* points */
	type->points = p = phgCalloc(3 * (o + 1), sizeof(*type->points));
	if (o == 0) {
	    p[0] = p[1] = p[2] = 0.5;
	}
	else {
	    /* TODO: use Gauss-Lobatto points */
	    d = 1.0 / o;
	    for (j = 0; j <= o; j++)
		p[j] = p[o + 1 + j] = p[2 * (o + 1) + j] = j * d;
	}
	type->work = phgCalloc(3 * (o + 1), sizeof(*type->work));
	/* orders */
	type->orders = phgCalloc(3, sizeof(*type->orders));
	type->orders[0] = type->orders[1] = type->orders[2] = o;
	type->dim = 1;			/* dim */
	type->nbas = (o + 1) * (o + 1) * (o + 1);	/* nbas */
	type->order = o /*+ o + o*/;		/* order */
	type->fe_space = FE_L2;		/* FE_SPACE */
	/* force initialization of type->work */
	for (j = 0; j <= o; j++) {
	    basis1D(o, j, 0.0, p,		type->work + j);
	    basis1D(o, j, 0.0, p + o + 1,	type->work + o + 1 + j);
	    basis1D(o, j, 0.0, p + 2 * (o + 1), type->work + 2 * (o + 1) + j);
	}

	/* DGPn */
	DOF_DGPn[o] = type = phgCalloc(1, sizeof(*type));
	sprintf(name, "DGP%d", o);
	type->name = strdup(name);	/* name */
	type->grad_type = DOF_DGPn[o == 0 ? o : o - 1]; /* grad_type */
	/* points */
	type->points = NULL;
	/* orders */
	type->orders = NULL;
	type->dim = 1;			/* dim */
	type->nbas = (o + 1) * (o + 2) * (o + 3) / 6;	/* nbas */
	type->order = o;		/* order */
	type->fe_space = FE_L2;		/* FE_SPACE */
	/* The basis functions are monomials x^i * y^j * z^k, i + j + k <= o,
	 * they are order as (here o := type->order):
	 * 	for (k = 0; k <= o; k++)
	 * 	    for (j = 0; j <= o - k; j++)
	 * 		for (i = 0; i <= o - k - j; i++)
	 * We store the powers for the l-th basis function in type->work[l]
	 * 	i + (j + k * (o + 1)) * (o + 1)
	 */
	type->work = phgAlloc(type->nbas * sizeof(*type->work));
	for (l = k = 0; k <= o; k++)
	    for (j = 0; j <= o - k; j++)
	 	for (i = 0; i <= o - k - j; i++)
		    type->work[l++] = i + (j + k * (o + 1)) * (o + 1);
	assert(l == type->nbas);
    }
    DOF_DEFAULT = DOF_DGPn[1];
}

static void
DOF_TYPE_free(void)
{
    int i, n = DOF_TYPE_ORDER_MAX + 1;
    DOF_TYPE *type;

    for (i = 0; i < n; i++) {
	type = DOF_DGQn[i];
	phgFree(type->name);
	phgFree(type->points);
	phgFree(type->work);
	phgFree(type->orders);
	phgFree(type);

	type = DOF_DGPn[i];
	phgFree(type->name);
	phgFree(type->work);
	phgFree(type);
    }
    phgFree(DOF_DGQn);
    phgFree(DOF_DGPn);
    DOF_DGQn = DOF_DGPn = NULL;
    DOF_DEFAULT = NULL;
}

#define ElementIsLocal(g, e) ((e)->index < (g)->nelem && (e)->index >= 0)

GRID *
phgNewGrid(int flags)
{
    GRID *g;
    Unused(flags);
    g = phgCalloc(1, sizeof(*g));
    return g;
}

static void
log_handler(FILE *log_stream, const char *filename, int lineno,
            int package, int category, int priority, const char *msg)
{
    /* do nothing, just to get rid of the annoying messages from p4est */
}

static const char *
default_dof_handler(OPTION *o, const char *arg, int prefix)
{
    char s0[128];

    Unused(prefix);

    if (arg == NULL) {
	/* print help message */
	phgOptionsPrintfHelp(o, "  List of available DOF types:");
	{
	    sprintf(s0, "    DGQ0...DGQ%d", DOF_TYPE_ORDER_MAX);
	    phgOptionsPrintfHelp(o, s0);
	}
    }
    else if (arg[0] != '\0') {	/* Note the convention with arg[0]=='\0' */
	const char *s = arg;
	if (s[0] == 'D' && s[1] == 'G' && isdigit(arg[2])) {
	    /* use "DGP" for "DG" */
	    strcpy(s0, "DGP");
	    strcpy(s0 + strlen(s0), s + 2);
	    s = s0;
	}
	if (!strncmp(s, "DGQ", 3) && isdigit(s[3])) {
	    int o = atoi(s + 3);
	    if (o < 0 || o > DOF_TYPE_ORDER_MAX)
		return NULL;
	    DOF_DEFAULT = DOF_DGQn[o];
	}
	else if (!strncmp(s, "DGP", 3) && isdigit(s[3])) {
	    int o = atoi(s + 3);
	    if (o < 0 || o > DOF_TYPE_ORDER_MAX)
		return NULL;
	    DOF_DEFAULT = DOF_DGPn[o];
	}
	else {
	    return NULL;
	}
    }

    return DOF_DEFAULT->name;
}

static void
modify_options()
{
    phgOptionsModify("-dof_type", "Default DOF type", NULL, default_dof_handler,
		     VT_HANDLER, FALSE);
}

void
phgInit(int *argc, char ***argv)
{
#undef phgInit
    extern void (*_phg_modify_options_callback)(void);	/* in option.c */

    DOF_TYPE_init();
    _phg_modify_options_callback = modify_options;
    phgInit(argc, argv);
    if (phgVerbosity > 0)
	phgPrintf("sizeof(ELEMENT) = %d\n", (int)sizeof(ELEMENT));
    sc_init(phgComm, 0, 0, NULL,
	    SC_LP_SILENT/*|SC_LP_PRODUCTION|SC_LP_ESSENTIAL*/);
    sc_set_log_defaults(NULL, log_handler, 0);
}

void
phgFinalize(void)
{
    DOF_TYPE_free();
    sc_finalize();
#undef phgFinalize
    phgFinalize();
}

FLOAT (*
phgGeomGetCuboid(GRID *g, ELEMENT *e, FLOAT (*box)[Dim]))[Dim]
{
    static FLOAT cuboid[2][Dim];
#if USE_OMP
# pragma omp threadprivate(cuboid)
#endif
    p8est_quadrant_t node;

    if (box == NULL)
	box = cuboid;

    p8est_quadrant_corner_node(e->quad, 0, &node);
    p8est_qcoord_to_vertex(g->conn, e->tree_id, node.x, node.y, node.z, box[0]);

    p8est_quadrant_corner_node(e->quad, 7, &node);
    p8est_qcoord_to_vertex(g->conn, e->tree_id, node.x, node.y, node.z, box[1]);

    return box;
}

static void
cb_elem(p8est_iter_volume_info_t *info, void *user_data)
{
    GRID *g = user_data;
    p8est_quadrant_t *quad = info->quad;
    ELEMENT *e = quad->p.user_data;

    e->index = g->i;
    e->tree = p8est_tree_array_index(g->p8est->trees, info->treeid);
    e->tree_id = info->treeid;
    e->quad = quad;
    phgGeomGetCuboid(g, e, e->cuboid);
    g->elems[g->i++] = e;
}

static void
update_elems(GRID *g)
{
    int i;

    g->nelem = g->p8est->local_num_quadrants;
    g->nelem_global = g->p8est->global_num_quadrants;

    phgFree(g->elems);
    g->elems = phgCalloc(g->nelem, sizeof(*g->elems));

    g->i = 0;
    p8est_iterate(g->p8est, g->ghost, /*user_data=*/g, cb_elem, NULL, NULL, NULL);
    assert(g->i == g->nelem);

    if (g->partition == NULL)
	g->partition = phgAlloc((g->nprocs + 1) * sizeof(*g->partition));
    MPI_Allgather(&g->nelem, 1, PHG_MPI_INT, g->partition + 1, 1, PHG_MPI_INT,
		  g->comm);
    g->partition[0] = 0;
    for (i = 0; i < g->nprocs; i++)
	g->partition[i + 1] += g->partition[i];
}

static void
setup_ghost(GRID *g)
{
    INT i;

    g->ghost = p8est_ghost_new (g->p8est, P8EST_CONNECT_FACE);
    g->ghost_data = phgAlloc(sizeof(ELEMENT) * g->ghost->ghosts.elem_count);

    /* FIXME: temporarily change e->index to global before ghost exchange */
    for (i = 0; i < g->nelem; i++)
	g->elems[i]->index += g->p8est->global_first_quadrant[g->rank];
    p8est_ghost_exchange_data(g->p8est, g->ghost, g->ghost_data);
    /* restore e->index to local */
    for (i = 0; i < g->nelem; i++)
	g->elems[i]->index -= g->p8est->global_first_quadrant[g->rank];

    g->mesh = p8est_mesh_new/*_ext*/(g->p8est, g->ghost, P8EST_CONNECT_FACE);

    if ((g->nghost = g->ghost->ghosts.elem_count) == 0)
	return;

    g->elems = phgRealloc0(g->elems, (g->nghost+g->nelem) * sizeof(*g->elems));
    g->O2Gmap = phgCalloc(g->nghost, sizeof(*g->O2Gmap));
    /* update user_data members of ghost elements */
    for (i = 0; i < g->nghost; i++) {
	p8est_quadrant_t *q = p8est_quadrant_array_index(&g->ghost->ghosts, i);
	ELEMENT *e = ((ELEMENT *)g->ghost_data) + i;
	e->quad = q;
	e->tree = NULL;
	g->O2Gmap[i] = e->index;
	e->index = g->nelem + i;
	g->elems[e->index] = e;
    }
}

static void
remove_ghost(GRID *g)
{
    if (g->mesh != NULL)
	p8est_mesh_destroy(g->mesh);
    g->mesh = NULL;
    if (g->ghost != NULL)
	p8est_ghost_destroy (g->ghost);
    g->ghost = NULL;
    phgFree(g->ghost_data);
    g->ghost_data = NULL;
    phgFree(g->O2Gmap);
    g->O2Gmap = NULL;
    g->nghost = 0;
}

void
phgFreeGrid(GRID **gptr)
{
    if (gptr == NULL || *gptr == NULL)
	return;
    remove_ghost(*gptr);
    p8est_destroy((*gptr)->p8est);
    p8est_connectivity_destroy((*gptr)->conn);
    phgFree((*gptr)->elems);
    phgFree((*gptr)->partition);
    phgFree(*gptr);
    *gptr = NULL;
}

BOOLEAN
phgImport(GRID *g, const char *filename, BOOLEAN distr)
{
    /* FIXME: import mesh from the given file */
    g->conn = p8est_connectivity_new_brick(1, 1, 1, 0, 0, 0);
    g->p8est = p8est_new_ext(phgComm, g->conn, 0, 0, 1, sizeof(ELEMENT), 
				NULL, NULL);
    g->nprocs = phgNProcs;
    g->rank = phgRank;
    g->comm = phgComm;
    if (distr) {
	p8est_partition(g->p8est, 0, NULL);
	g->balanced = TRUE;
    }
    update_elems(g);
    return TRUE;
}

static int
cb_refine(p8est_t *p8est, p8est_topidx_t which_tree, p8est_quadrant_t *quad)
{
    return 1;
}

void
phgRefineAllElements(GRID *g, int level)
{
    if (g->ghost != NULL)
	remove_ghost(g);
    while (level-- > 0)
	p8est_refine(g->p8est, /*recursive*/0, cb_refine, NULL);
    p8est_balance(g->p8est, P8EST_CONNECT_FACE, NULL);
    g->balanced = FALSE;
    update_elems(g);
}

int
phgBalanceGrid(GRID *g, FLOAT lif_threshold, INT submesh_threshold,
		DOF *weights, FLOAT power)
{
    if (!g->balanced) {
	if (g->ghost != NULL)
	    remove_ghost(g);
	p8est_partition(g->p8est, 0, NULL);
	update_elems(g);
	g->balanced = TRUE;
    }
    return 1;
}

void
phgSetupHalo_(GRID *g, int halo_type, BOOLEAN update_dofs)
{
    setup_ghost(g);

#if 0	/* for debugging */
#warning debugging code!
    phgInfo(-1, "%d local elements:\n", g->nelem);
    for (INT i = 0; i < g->nelem; i++) {
	ELEMENT *e = g->elems[i];
	phgInfo(-1, "elem %d: box=(%lg,%lg,%lg)-(%lg,%lg,%lg)\n",
			GlobalElement(g, e->index), 
			e->cuboid[0][0], e->cuboid[0][1], e->cuboid[0][2],
			e->cuboid[1][0], e->cuboid[1][1], e->cuboid[1][2]);
	for (int j = 0; j < NFace; j++) {
	    ELEMENT *e1 = phgGetNeighbour(g, e, j);
	    if (e1 == NULL || e1->index < 0)
		continue;
	    phgInfo(-1, "  neighbor[%d]=%d: box=(%lg,%lg,%lg)-(%lg,%lg,%lg)\n",
			j, GlobalElement(g, e1->index),
			e1->cuboid[0][0], e1->cuboid[0][1], e1->cuboid[0][2],
			e1->cuboid[1][0], e1->cuboid[1][1], e1->cuboid[1][2]);
	}
    }
#endif
}

size_t
phgMemoryUsage(GRID *g, size_t *peak)
{
    /* FIXME: modify phgMemoryUsage1 to use MPI_Comm instead of GRID */
#ifdef phgMemoryUsage
# undef phgMemoryUsage	/* we need the original phgMemoryUsage of PHG here */
#endif
    return phgMemoryUsage(NULL, peak);
}

void
phgGeomGetFaceOutNormal(GRID *g, ELEMENT *e, int face, FLOAT *nv)
/* Note: the faces are ordered as -x, +x, -y, +y, -z, +z */
{
    switch (face) {
	case 0:	nv[0] =-1.; nv[1] = 0.; nv[2] = 0.; break;
	case 1: nv[0] = 1.; nv[1] = 0.; nv[2] = 0.; break;
	case 2: nv[0] = 0.; nv[1] =-1.; nv[2] = 0.; break;
	case 3: nv[0] = 0.; nv[1] = 1.; nv[2] = 0.; break;
	case 4: nv[0] = 0.; nv[1] = 0.; nv[2] =-1.; break;
	case 5: nv[0] = 0.; nv[1] = 0.; nv[2] = 1.; break;
	default: phgError(1, "%s: invalid face no. (%d)\n", __func__, face);
    }
}

ELEMENT *
phgGetNeighbour_(GRID *g, ELEMENT *e, int face, const char *file, int line)
/* Note: the faces are ordered as -x, +x, -y, +y, -z, +z.
 *
 * See p4est_mesh_face_neighbor_next() (src/p4est_mesh.c) in the source files
 * of p4est for information about related structs */
{
    int q2f;
    p8est_locidx_t q2q, quadfacecode;
    ELEMENT *e1;

    /* e must be local! */
    if (!ElementIsLocal(g, e)) {
	phgError(1, "%s:%d: unexpected error!\n", file, line);
    }

    assert(g->ghost != NULL);
    assert(face >= 0 && face < P8EST_FACES);
    quadfacecode = P8EST_FACES * (e->tree->quadrants_offset +
		   	sc_array_position(&e->tree->quadrants, e->quad)) + face;
    q2q = g->mesh->quad_to_quad[quadfacecode];
    q2f = (int)g->mesh->quad_to_face[quadfacecode];
    if (q2f < 0) {
	/* neighbour is half size, return a dummy element with index -1 */
	static ELEMENT dummy = {{{0.,0.,0.},{0.,0.,0.}}, /* cuboid */
				NULL,	/* tree */
				NULL,	/* quad */
				-1,	/* index */
				-1,	/* tree_id */
				0	/* mark */};
	return &dummy;
    }
    if (q2q < g->mesh->local_num_quadrants) {
	p8est_quadrant_t *q;
	q = p8est_mesh_quadrant_cumulative(g->p8est, q2q, NULL, NULL);
	e1 = q->p.user_data;
    }
    else {
	q2q -= g->mesh->local_num_quadrants;
	assert(q2q < g->mesh->ghost_num_quadrants);
	e1 = ((ELEMENT *)g->ghost_data) + q2q;
    }
    return e1 == e ? NULL : e1;
}

BYTE
phgOppositeFace(GRID *g, ELEMENT *e, BYTE v, ELEMENT *e_op)
{
    int q2f;
    p8est_locidx_t quadfacecode;

    /* The code below only allows e to be local! */
    assert(ElementIsLocal(g, e));
    assert(v >= 0 && v < P8EST_FACES);
    quadfacecode = P8EST_FACES * (e->tree->quadrants_offset +
		   	sc_array_position(&e->tree->quadrants, e->quad)) + v;
    q2f = g->mesh->quad_to_face[quadfacecode];
    assert(q2f >= 0);
    return q2f % 6;
}

DOF *
phgDofNew_(GRID *g, DOF_TYPE *type, HP_TYPE *hp, SHORT dim,
	   const char *name, DOF_USER_FUNC userfunc,
	   const char *srcfile, int srcline)
{
    DOF *dof = phgCalloc(1, sizeof(*dof));

    assert(hp == NULL);
    dof->g = g;
    dof->dim = dim;
    dof->type = type;
    dof->name = name == NULL ? NULL : strdup(name);
    dof->userfunc = (void *)userfunc;
    dof->poly_order = -1;

    if (dof->type != DOF_ANALYTIC) {
	int nbas = DofNBas(dof, NULL);
	dof->data = phgCalloc(g->nelem, nbas * dim * sizeof(FLOAT));
	dof->ghost_data = phgCalloc(g->nghost, nbas * dim * sizeof(FLOAT));
	dof->poly_order = DofTypeOrder(dof, NULL);
	if (!SpecialDofUserFunction(userfunc))
	    phgDofSetDataByFunction(dof, userfunc);
    }

    return dof;
}

static void
dof_update_ghost(DOF *dof)
/* get DOF ghost data (see p4est_ghost.c/p4est_ghost_exchange_data_begin)
 * FIXME: allow multiple DOFs, for better communication performance */
{
    p8est_ghost_t *ghost = dof->g->ghost;
    void **mirror_data = phgAlloc(ghost->mirrors.elem_count * sizeof(void *));
    int data_count = dof->dim * DofNBas(dof, NULL);	/* FLOATs/element */
    size_t data_size = data_count * sizeof(*dof->data);
    p8est_quadrant_t *mirror, *q;
    p8est_topidx_t which_tree;
    p8est_locidx_t which_quad;
    p8est_tree_t *tree;
    ELEMENT *e;
    INT i;

    for (i = 0; i < ghost->mirrors.elem_count; ++i) {
	mirror = p8est_quadrant_array_index(&ghost->mirrors, i);
	which_tree = mirror->p.piggy3.which_tree;
	tree = p8est_tree_array_index (dof->g->p8est->trees, which_tree);
	which_quad = mirror->p.piggy3.local_num - tree->quadrants_offset;
	q = p8est_quadrant_array_index (&tree->quadrants, which_quad);
	e = q->p.user_data;
	mirror_data[i] =dof->data + e->index * data_count;
    }
    p8est_ghost_exchange_custom(dof->g->p8est, ghost, data_size,
					mirror_data, dof->ghost_data);
    phgFree(mirror_data);
}

/* context used by the callback functions */
typedef struct {
    int i1, i2;			/* ints */
    void *p1, *p2, *p3, *p4;	/* pointers */
} CTX;
static CTX *ctx = NULL;
#if USE_OMP
# pragma omp threadprivate(ctx)
#endif	/* USE_OMP */

static void
f_wrapper(FLOAT x, FLOAT y, FLOAT z, FLOAT *res)
{
    if (ctx->i1 == 0) {
	/* DOF_USER_FUNC */
	((DOF_USER_FUNC)ctx->p1)(x, y, z, res);
    }
    else {
	/* DOF_USER_FUNC_LAMBDA */
	FLOAT xref[3], xyz[3] = {x, y, z};
	DOF *u = ctx->p2;
	ELEMENT *e = ctx->p3;
	xyz_to_ref(u->g, e->index, xyz, xref);
	((DOF_USER_FUNC_LAMBDA)ctx->p1)(u, e, 0, xref, res); 
    }
}

void
phgDofSetDataByFunction_(DOF *dof, DOF_USER_FUNC func,
		DOF_USER_FUNC_LAMBDA func_lam, FLOAT *funcvalues)
/* compute DOFs by L2 projection or interpolation in each element */
{
    GRID *g = dof->g;
    FLOAT *data = dof->data;
    INT i;
    QCACHE *qc = phgQCNew(QD_P4EST, dof);
    int Q_f;
    CTX *ctx_old;

    assert(funcvalues == NULL);		/* unimplemented */

    if (func == NULL && func_lam == NULL) {
	size_t size = DofNBas(dof, NULL) * dof->dim * sizeof(FLOAT);
	memset(dof->data, 0, g->nelem * size);
	memset(dof->ghost_data, 0, g->nghost * size);
	return;
    }

    /* Note: implement phgQCIntegrate functions returning multiple FLOATs */
    ctx_old = ctx;
    ctx = phgCalloc(1, sizeof(*ctx));
    ctx->i1 = func != NULL ? 0 : 1;	/* 0: FUNC, 1: FUNC_LAMBDA */
    ctx->p1 = func != NULL ? (void *)func : (void *)func_lam;
    ctx->p2 = dof;
    Q_f = phgQCAddXYZFunction(qc, f_wrapper, dof->dim);
    for (i = 0; i < g->nelem; i++) {
	ELEMENT *e = g->elems[i];
	int j, n = DofNBas(dof, e);
	if (dof->type->points != NULL) {	/* Interpolation */
	    FLOAT x[Dim], tmp[Dim], *pt = tmp;
	    for (j = 0; j < n; j++, data += dof->dim) {
		if (dof->type->orders == NULL) {
		    pt = dof->type->points + j * Dim;
		}
		else  {	/* j = jx + (jy + jz * (orders[1]+1)) * (orders[0]+1) */
		    int jx, jy, jz;
		    jx = j % (dof->type->orders[0] + 1);
		    jz = j / (dof->type->orders[0] + 1);
		    jy = jz % (dof->type->orders[1] + 1);
		    jz /= (dof->type->orders[1] + 1);
		    tmp[0] = dof->type->points[jx];
		    tmp[1] = dof->type->points[dof->type->orders[0] + 1 + jy];
		    tmp[2] = dof->type->points[dof->type->orders[0] + 1 +
					       dof->type->orders[1] + 1 + jz];
		}
		if (func == NULL) {
		    func_lam(dof, e, 0, pt, data);
		    continue;
		}
		ref_to_xyz(dof->g, e->index, pt, x);
		func(x[0], x[1], x[2], data);
	    }
	}
	else {					/* elementwise L2 projection */
	    FLOAT (*cuboid)[Dim] = phgGeomGetCuboid(g, e, NULL);
	    int k, pvt[n];
	    /* FIXME: use projection on the reference element, save LU of A */
	    FLOAT A[n][n], b[dof->dim][n], buffer[dof->dim];
	    FLOAT *rule;
	    phgQuadInterfaceCuboid(NULL, 0, NULL, cuboid, dof->poly_order * 2,
				   &rule, NULL, NULL);
	    phgQCSetRule(qc, rule, -1.0);
	    ctx->p3 = e;
	    for (j = 0; j < n; j++) {
		for (k = 0; k <= j; k++)
		    A[j][k] = A[k][j] = 
			phgQCIntegrate(qc, i, Q_BAS, j, qc, i, Q_BAS, k);
/* FIXME: without one of the two lines below interface-p4est segfaults.
 *  2021.05.01: it might be a GCC bug on Fedora 33, with gcc version:
 *			10.3.1 20210422 (Red Hat 10.3.1-1) (GCC)
 *  2010.05.05: it seems the bug doesn't show up if QCINTEGRATE_HACK is defined
 *		(see quad-cache.h) */
#ifndef QCINTEGRATE_HACK
//bzero(buffer, sizeof(buffer));
buffer[0] = *
#endif
		phgQCIntegrateM(qc, i, Q_BAS, j, qc, i, Q_f, 0,
				/*N*/1, /*M*/dof->dim, /*K*/1, buffer);
		for (k = 0; k < dof->dim; k++)
		    b[k][j] = buffer[k];
	    }
	    phgFree(rule);
	    phgSolverDenseLU(n, (FLOAT *)A, pvt);
	    phgSolverDenseSV(n, (FLOAT *)A, pvt, dof->dim, (FLOAT *)b);
	    for (j = 0; j < n; j++)
		for (k = 0; k < dof->dim; k++)
		    *(data++) = b[k][j];
	}
    }
    phgFree(ctx);
    ctx = ctx_old;

    phgQCFree(&qc);
    dof_update_ghost(dof);
}

void
phgDofSetPolyOrder(DOF *u, SHORT order)
{
    u->poly_order = order;
}

void
phgDofFree(DOF **pdof)
{
    DOF *dof;

    if (pdof == NULL || *pdof == NULL)
	return;
    dof = *pdof;
    *pdof = NULL;
    phgFree(dof->name);
    phgFree(dof->data);
    phgFree(dof->ghost_data);
    phgFree(dof);
}

FLOAT *
phgDofEval(DOF *dof, ELEMENT *e, const FLOAT xref[], FLOAT *values)
{
    int i, j, nb;
    FLOAT bas, *data;

    if (dof->type == DOF_ANALYTIC) {
	FLOAT x[3];
	ref_to_xyz(dof->g, e->index, (FLOAT *)xref, x);
	dof->userfunc(x[0], x[1], x[2], values);
	return values;
    }

    nb = DofNBas(dof, e);
    if (e->index < dof->g->nelem)
	data = dof->data + dof->dim * (size_t)e->index * nb;
    else
	data = dof->ghost_data +
			dof->dim * (size_t)(e->index - dof->g->nelem) * nb;
    for (i = 0; i < dof->dim; i++)
	values[i] = 0.;
    for (j = 0; j < nb; j++) {
	basisFunc(dof, xref[0], xref[1], xref[2], j, &bas);
	for (i = 0; i < dof->dim; i++)
	    values[i] += *(data++) * bas;
    }

    return values;
}

DOF *
phgDofAXPBY_(FLOAT a, DOF *x, FLOAT b, DOF **py,
	     const char *srcfile, int srcline, BOOLEAN check)
{
    size_t i;
    DOF *y = py == NULL ? NULL : *py;
    int nbas = DofNBas(x, NULL);

    if (y == NULL)
	y = phgDofNew(x->g, x->type, x->dim, "y", DofNoAction);
    assert(x->type == y->type && x->dim == y->dim && x->g == y->g &&
	   x->type != DOF_ANALYTIC);
    for (i = 0; i < x->g->nelem * (size_t)nbas * x->dim; i++)
	y->data[i] = a * x->data[i] + b * y->data[i];
    for (i = 0; i < x->g->nghost * (size_t)nbas * x->dim; i++)
	y->ghost_data[i] = a * x->ghost_data[i] + b * y->ghost_data[i];
    if (py != NULL)
	*py = y;

    return y;
}

static DOF *grad_src = NULL;
#if USE_OMP
# pragma omp threadprivate(grad_src)
#endif	/* USE_OMP */
static void
cb_grad(DOF *u, ELEMENT *e, int bno, const FLOAT xref[], FLOAT *grad)
/* callback function for phgDofGradient_ */
{
    FLOAT dx[Dim], *data, (*cuboid)[Dim];
    int i, j, n;

    u = grad_src;	/* the src DOF */
    n = DofNBas(u, NULL);
    data = u->data + e->index * n * u->dim;
    cuboid = /*phgGeomGetCuboid(u->g, e, NULL)*/e->cuboid;
    dx[0] = 1.0 / (cuboid[1][0] - cuboid[0][0]);
    dx[1] = 1.0 / (cuboid[1][1] - cuboid[0][1]);
    dx[2] = 1.0 / (cuboid[1][2] - cuboid[0][2]);
    /* data[n][dim] -> grad[dim][Dim] */
    for (j = 0; j < u->dim * Dim; j++)
	grad[j] = 0.;
    for (i = 0; i < n; i++, data += u->dim) {
	FLOAT grad0[Dim];
	gradBasisFunc(u, xref[0], xref[1], xref[2], i, grad0);
	for (j = 0; j < u->dim; j++) {
	   grad[j * Dim + 0] += grad0[0] * dx[0] * data[j];
	   grad[j * Dim + 1] += grad0[1] * dx[1] * data[j];
	   grad[j * Dim + 2] += grad0[2] * dx[2] * data[j];
	}
    }
}

DOF *
phgDofGradient_(DOF *src, DOF **newdof, DOF_TYPE *newtype,
		const char *name, const char *srcfile, int srcline)
{
    DOF *grad = NULL;

    assert(src->type != DOF_ANALYTIC);
    if (newtype == NULL)
	newtype = src->type->grad_type;
    /*assert(newtype == src->type);*/

    if (newdof != NULL)
	grad = *newdof;
    if (grad != NULL)
	phgDofFree(&grad);
    grad = phgDofNew(src->g, newtype, src->dim * Dim, "grad", DofNoAction);
    grad_src = src;	/* for use in cb_grad */
    phgDofSetDataByLambdaFunction(grad, cb_grad);
    if (newdof != NULL)
	*newdof = grad;

    return grad;
}

FLOAT
phgDofNormL2(DOF *u)
{
    ELEMENT *e;
    FLOAT res;
    QCACHE *qc;
    int Q_u;

    qc = phgQCNew(QD_P4EST, u);
    Q_u = phgQCAddFEFunction(qc, u);
    res = 0.;
#if USE_OMP
# pragma omp parallel for private(e) reduction(+:res)
#endif	/* USE_OMP */
    ForAllElementsBegin(u->g, e)
	FLOAT *rule;
	int o = 2 * DofTypeOrder(u, e);
	if (o < 0)
	    o = 2;
	phgQuadInterfaceCuboid(NULL, 0, NULL, e->cuboid, o, &rule, NULL, NULL);
	phgQCSetRule(qc, rule, -1.0);
	res += phgQCIntegrate(qc, e->index, Q_u, 0, qc, e->index, Q_u, 0);
	phgFree(rule);
    ForAllElementsEnd
    phgQCFree(&qc);

#if USE_MPI
    if (u->g->nprocs > 1) {
	FLOAT res0 = res;
	MPI_Allreduce(&res0, &res, 1, PHG_MPI_FLOAT, PHG_SUM, u->g->comm);
    }
#endif	/* USE_MPI */

    return Sqrt(res);
}

INT
phgQuadInterfaceMarkGrid(DOF *ls, DOF *ls_grad)
{
    GRID *g = ls->g;
    ELEMENT *e;
    INT n0 = 0;

    assert(ls->type == DOF_ANALYTIC && ls_grad->type == DOF_ANALYTIC);

#if USE_OMP
#pragma omp parallel for private(e) reduction(+:n0)
#endif	/* USE_OMP */
    ForAllElementsBegin(g, e)
	FLOAT (*cuboid)[Dim] = phgGeomGetCuboid(g, e, NULL);
	int ls_order = DofTypeOrder(ls, e);
	e->mark = phgQuadInterfaceMarkElement(ls->userfunc, ls_order,
					ls_grad->userfunc, ET_CUBOID, cuboid);
	if (e->mark == 0)
	    n0++;
    ForAllElementsEnd
 
    return n0;
}

FLOAT
phgGeomGetDiameter(GRID *g, ELEMENT *e)
{
    FLOAT (*box)[Dim] = phgGeomGetCuboid(g, e, NULL);
    FLOAT dx = box[1][0] - box[0][0],
	  dy = box[1][1] - box[0][1],
	  dz = box[1][2] - box[0][2];

    return Sqrt(dx * dx + dy * dy + dz * dz);
}

FLOAT
phgGeomGetFaceDiameter(GRID *g, ELEMENT *e, int face)
{
    FLOAT (*box)[Dim] = phgGeomGetCuboid(g, e, NULL);
    FLOAT dx, dy, dz;

    assert(face >= 0 && face < P8EST_FACES);
    face /= 2;
    dx = face == 0 ? 0. : box[1][0] - box[0][0];
    dy = face == 1 ? 0. : box[1][1] - box[0][1];
    dz = face == 2 ? 0. : box[1][2] - box[0][2];

    return Sqrt(dx * dx + dy * dy + dz * dz);
}

FLOAT
phgGeomGetVolume(GRID *g, ELEMENT *e)
{
    FLOAT (*box)[Dim] = phgGeomGetCuboid(g, e, NULL);

    return (box[1][0] - box[0][0]) *
	   (box[1][1] - box[0][1]) *
	   (box[1][2] - box[0][2]);
}

INT
phgMapE2G(MAP *map, int dof_id, ELEMENT *e, int index)
{
    DOF *u;
    GRID *g;
    INT no, nelem;
    int rank;

    assert(dof_id >= 0 && dof_id < map->ndof);
    u = (DOF *)map->dofs[dof_id];
    g = u->g;
    if (ElementIsLocal(u->g, e)) {
	no = u->dim * e->index * DofNBas(u, e) + index;
	rank = g->rank;
	nelem = g->nelem;
    }
    else {
	/* find owner process of e by binary search */
	no = g->O2Gmap[e->index - g->nelem];
	rank = (int)phgBinarySearchINT(g->nprocs, g->partition, NULL, no);
	if (no != g->partition[rank])
	    rank--;
	assert(no >= g->partition[rank] && no < g->partition[rank + 1]);
	no = u->dim * (no - g->partition[rank]) * DofNBas(u, e) + index;
	nelem = g->partition[rank + 1] - g->partition[rank];
    }

    while (dof_id > 0) {
	u = (DOF *)map->dofs[--dof_id];
	no += u->dim * nelem * DofNBas(u, e);
    }

    return no + map->partition[rank];
}

#include <stdarg.h>

const char *
phgExportVTK(GRID *g, const char *fn, DOF *dof, ...)
{
    /* FIXME: include DOFs in the VTK file. */
    p8est_vtk_write_file(g->p8est, NULL, fn);
    return fn;
}

SOLVER *
phgSolverCreate(OEM_SOLVER *oem_solver, DOF *u, ...)
{
    GRID *g = u->g;
    MAP *map;
    MAT *mat;
    INT n, N;
    int i, ndof, nb = DofNBas(u, NULL);
    DOF **dofs;
    SOLVER *solver;

    GetVariableArgs(ndof, dofs, u, DOF *);

    /* count # local unkowns */
    n = 0;
    for (i = 0; i < ndof; i++) {
	assert(dofs[i]->type != DOF_ANALYTIC);
	n += dofs[i]->dim * g->nelem * nb;
    }
    N = n;
#if USE_MPI
    if (g->nprocs > 1)
	MPI_Allreduce(&n, &N, 1, PHG_MPI_INT, MPI_SUM, g->comm);
#endif	/* USE_MPI */

    /* create MAP */
    map = phgMapCreateSimpleMap(g->comm, n, N);
    /* save DOFs list to map->dofs */
    phgFree(map->dofs);
    map->ndof = ndof;
    map->dofs = (void *)dofs;

    /* create MAT */
    mat = phgMapCreateMat(map, map);
    phgMapDestroy(&map);
    mat->handle_bdry_eqns = FALSE;

    /* create SOLVER */
    solver = phgSolverMat2Solver(oem_solver, mat);
    phgMatDestroy(&mat);
    phgVecDisassemble(solver->rhs);
    solver->rhs_updated = TRUE;

    return solver;
}

int
phgSolverSolve(SOLVER *solver, BOOLEAN destroy, DOF *u, ...)
{
    GRID *g = u->g;
    DOF **dofs;
    VEC *v = phgMapCreateVec(solver->rhs->map, 1);
    FLOAT *data;
    int i, ndof;
    INT n;

    GetVariableArgs(ndof, dofs, u, DOF *);

    /* copy DOF to v as initial solution */
    data = v->data;
    for (i = 0; i < ndof; i++) {
	n = dofs[i]->dim * g->nelem * DofNBas(dofs[i], NULL);
	memcpy(data, dofs[i]->data, n * sizeof(*data));
	data += n;
    }

    phgSolverVecSolve(solver, destroy, v);

    /* copy solution in v to DOF */
    data = v->data;
    for (i = 0; i < ndof; i++) {
	n = dofs[i]->dim * g->nelem * DofNBas(dofs[i], NULL);
	memcpy(dofs[i]->data, data, n * sizeof(*data));
	data += n;
	dof_update_ghost(dofs[i]);
    }

    phgVecDestroy(&v);

    phgFree(dofs);

    return 0;
}

void
phgQuadInterface(DOF *ls, DOF *ls_grad, ELEMENT *e, int order,
		 FLOAT **rule_m, FLOAT **rule_0, FLOAT **rule_p)
{
   GRID *g;
   FUNC_3D ls_func, grad_func;
   int ls_order;

   assert(ls != NULL || ls_grad != NULL);
   g = (ls != NULL ? ls : ls_grad)->g;
   ls_order = (ls != NULL ? ls : ls_grad)->poly_order;
   ls_func = ls != NULL ? ls->userfunc : NULL;
   grad_func = ls_grad != NULL ? ls_grad->userfunc : NULL;

   phgQuadInterfaceCuboid(ls_func, ls_order, grad_func,
			  phgGeomGetCuboid(g, e, NULL), order,
			  rule_m, rule_0, rule_p);
}

void
phgQuadInterfaceFace(DOF *ls, DOF *ls_grad, ELEMENT *e, int face, int order,
		     FLOAT **rule_m, FLOAT **rule_0, FLOAT **rule_p)
{
   GRID *g;
   FUNC_3D ls_func, grad_func;
   int ls_order;

   assert(ls != NULL || ls_grad != NULL);
   g = (ls != NULL ? ls : ls_grad)->g;
   ls_order = (ls != NULL ? ls : ls_grad)->poly_order;
   ls_func = ls != NULL ? ls->userfunc : NULL;
   grad_func = ls_grad != NULL ? ls_grad->userfunc : NULL;

   phgQuadInterfaceCuboidFace(ls_func, ls_order, grad_func,
			      phgGeomGetCuboid(g, e, NULL), face, order,
			      rule_m, rule_0, rule_p);
}

FLOAT *
phgQuadGetRule_(GRID *g, ELEMENT *e, int which, int order)
/* returns a rule of order "order" for integration in the element "e", or
 * one of its face or edge. "which" specifies which part of the element is 
 * the integration domain:
 * 		which = 3 * no + (dim - 1)
 * where "no" is the face ("dim" == 2) or edge ("dim" == 1) number. */
{
    int dim = which % 3 + 1, no = which / 3;
    FLOAT *rule = NULL;

    if (dim == 3)
	phgQuadInterfaceCuboid(NULL, 0, NULL, e->cuboid, order,
							&rule, NULL, NULL);
    else if (dim == 2)
	phgQuadInterfaceCuboidFace(NULL, 0, NULL, e->cuboid, no, order,
							&rule, NULL, NULL);
    else
	phgError(1, "%s: unimplemented case (dim == 1)\n");

    return rule;
}

/*----------------------- End of PHG_to_p8est wrapper -----------------------*/
