/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: halo.c,v 1.36 2021/08/11 13:38:49 zlb Exp $ */

#include "phg.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>

ELEMENT *
phgGetNeighbour_(GRID *g, ELEMENT *e, int face, const char *file, int line)
/* returns the neighbour element for interior face and NULL for boundary face.
 * Note: if the neighbour is in another submesh, then g->halo must be set up
 * before calling this function */
{
    if (g->halo != NULL && e->index >= g->halo->nelem_bak)
	phgError(1, "%s:%d: can't access neighbours of halo elements.\n",
			file, line);

    if (!HasRemoteNeighbour(e, face))
	return GetNeighbour(e, face);

    if (g->halo == NULL)
	phgError(1, "%s:%d: cannot access remote neighbours, please call "
		"phgSetupHalo before using this function.\n", file, line);

    return g->halo->elist + g->halo->map[(size_t)(e->neighbours[face])];
}

#if USE_MPI
/* static variables for searching sorted arrays L2Gmap[ordering[...]] */
static INT *L2Gmap, *ordering;

static int
comp_L2Gmap(const void *p0, const void *p1)
{
    INT i0 = L2Gmap[*(const INT *)p0], i1 = L2Gmap[*(const INT *)p1];
    return i0 == i1 ? 0 : (i0 < i1 ? -1 : 1);
}

/* static variables for sorting new items */
static struct {
    INT gindex, oindex, pindex;	/* global/owner/periodic index of the item */
    INT rbuf_index;		/* rbuf index: rbuf[rbuf_index].elem */
    int item_index, orank;	/* item index and owner rank, e.g.,
				 * rbuf[rbuf_index].elem.verts[item_index] */
    BTYPE type;			/* boundary type */
} *items, *pi;

/* struct for storing data from the remote neighbour
 * TODO: communications can be reduced w.r.t.:
 *   1. don't send entire ELEMENTs, only those fields actually needed.
 *   2. if g->period == NULL, then
 * 		pindex and {coord,indexes,ranks,types}[i] for i<4 and i!=vertex
 *	are not needed, which saves:
 *		3 "int" + 4 "INT" + 3 "COORD" + 3 "BTYPE" \approx 100 bytes
 */
static struct {
    ELEMENT	elem;		/* element, with e.index, e.verts[], e.edges[]
				 * and e.faces[] containing global indices */
    COORD	coord[4];	/* coordinates of the vertices */
    INT		pindex;		/* g->period->L2Gmap_vert for the opp vertex */
    INT		indexes[11];	/* owner indexes: V+E+F+E = 4+3+3+1 */
    int		ranks[11];	/* owner ranks:	V+E+F+E = 4+3+3+1 */
    BTYPE	types[11];	/* types: V+E+F+E = 4+3+3+1 */
} *sbuf, *rbuf;

static int
comp_item(const void *p0, const void *p1)
{
    INT i0 = items[*(const INT *)p0].gindex,
	i1 = items[*(const INT *)p1].gindex;
    return i0 == i1 ? 0 : (i0 < i1 ? -1 : 1);
}

static int
comp_elem(const void *p0, const void *p1)
{
    INT i0 = rbuf[*(const INT *)p0].elem.index,
	i1 = rbuf[*(const INT *)p1].elem.index;
    return i0 == i1 ? 0 : (i0 < i1 ? -1 : 1);
}

#endif	/* USE_MPI */

#define REALLOC(p, n, m) p = phgRealloc_(p, (n) * sizeof(*p), (m) * sizeof(*p))

void
phgSetupHalo_(GRID *g, int halo_type, BOOLEAN update_dofs)
/* If 'update_dofs' != FALSE, then previously defined DOF objects will be
 * updated with correct data for the halo elements unless userfunc==DofNoAction.
 *
 * The argument "halo_type" specifies the kind and size (width) of the halo,
 * see the macros HaloType, GetHaloSize, GetHaloKind, etc, in halo.h.
 */
{
#if USE_MPI
    INT i, j, k, n, m;
    int ii, jj, vertex, peer_vertex, u[3], v[3];
    ELEMENT *e, *peer;
    MPI_Datatype type;
    DOF **pdof;
    HP_TYPE **php;

    if (g->halo != NULL) {
	if (g->halo->type == halo_type)
	    return;
	phgRemoveHalo(g);
	assert(g->halo == NULL);
    }

    if (halo_type < 0)
	return;

    if (GetHaloKind(halo_type) != HALO_FACE || GetHaloSize(halo_type) > 1)
	phgError(1, "%s (%s:%d): unimplemented feature.\n",
				__func__, __FILE__, __LINE__);

    g->halo = phgCalloc(1, sizeof(*g->halo));
    g->halo->type = halo_type;
    g->halo->nvert_bak = g->nvert;
    g->halo->nedge_bak = g->nedge;
    g->halo->nface_bak = g->nface;
    g->halo->nelem_bak = g->nelem;

    if (g->nprocs == 1)
	return;

    /* invalidate face->element map */
    phgFree(g->Fmap_elems);
    g->Fmap_elems = NULL;
    phgFree(g->Fmap_faces);
    g->Fmap_faces = NULL;

    if (g->owner_index_face == NULL) {
	/* set up g->owner_{index,rank}_face */
	assert(g->owner_rank_face == NULL);
	g->owner_index_face = phgAlloc(g->nface * sizeof(*g->owner_index_face));
	g->owner_rank_face = phgAlloc(g->nface * sizeof(*g->owner_rank_face));
	for (i = 0; i < g->nface; i++) {
	    g->owner_index_face[i] = -1;
	    g->owner_rank_face[i] = -1;
	}
	for (i = 0; i < g->neighbours.count; i++) {
	    RNEIGHBOUR *rn = g->neighbours.list + i;
	    j = rn->local->faces[rn->vertex];
	    if ((g->types_face[j] & OWNER))
		continue;
	    g->owner_index_face[j] = rn->peer_face;
	    g->owner_rank_face[j] = rn->rank;
	}
    }

    if (g->owner_index_elem == NULL) {
	/* set up g->owner_{index,rank}_elem */
	assert(g->owner_rank_elem == NULL);
	g->owner_index_elem = phgAlloc(g->nface * sizeof(*g->owner_index_elem));
	g->owner_rank_elem = phgAlloc(g->nface * sizeof(*g->owner_rank_elem));
	for (i = 0; i < g->nelem; i++) {
	    if ((g->types_elem[i] & OWNER)) {
		g->owner_index_elem[i] = g->elems[i]->index;
		g->owner_rank_elem[i] = g->rank;
	    }
	    else {
		g->owner_index_elem[i] = -1;
		g->owner_rank_elem[i] = -1;
	    }
	}
    }

    /* prepare data to send */
    sbuf = phgAlloc(g->neighbours.count * sizeof(*sbuf));
    for (i = 0; i < g->neighbours.count; i++) {
	int kk;
	BTYPE t;
	e = g->neighbours.list[i].local;
	ii = g->neighbours.list[i].vertex;
	GetFaceVertices(e, ii, v[0], v[1], v[2], jj);
	/*----- copy period->L2Gmap_vert value of the opposite vertex */
	sbuf[i].pindex = GlobalVertexP(g, e->verts[ii]);
	/*----- copy coordinates and owner_index_*[] and owner_rank_*[] data */
	for (kk = 0; kk < 4; kk++) {
	    memcpy(sbuf[i].coord + kk, g->verts + e->verts[kk], sizeof(COORD));
	    sbuf[i].types[kk] = (t = g->types_vert[e->verts[kk]]) & ~OWNER;
	    if ((t & OWNER)) {
		/* Note: corresponding entries in g->owner_* are -1 */
		sbuf[i].indexes[kk] = e->verts[kk];
		sbuf[i].ranks[kk] = g->rank;
	    }
	    else {
		sbuf[i].indexes[kk] = g->owner_index_vert[e->verts[kk]];
		sbuf[i].ranks[kk] = g->owner_rank_vert[e->verts[kk]];
	    }
	}
	/* the three edges */
	for (kk = 0; kk < 3; kk++) {
	    jj = GetEdgeNo(ii, v[kk]);
	    sbuf[i].types[4 + kk] = (t = g->types_edge[e->edges[jj]]) & ~OWNER;
	    if ((t & OWNER)) {
		/* Note: corresponding entries in g->owner_* are -1 */
		sbuf[i].indexes[4 + kk] = e->edges[jj];
		sbuf[i].ranks[4 + kk] = g->rank;
	    }
	    else {
		sbuf[i].indexes[4 + kk] = g->owner_index_edge[e->edges[jj]];
		sbuf[i].ranks[4 + kk] = g->owner_rank_edge[e->edges[jj]];
	    }
	}
	/* the three faces */
	for (kk = 0; kk < 3; kk++) {
	    jj = v[kk];
	    sbuf[i].types[7 + kk] = (t = g->types_face[e->faces[jj]]) & ~OWNER;
	    if ((t & OWNER)) {
		/* Note: corresponding entries in g->owner_* are -1 */
		sbuf[i].indexes[7 + kk] = e->faces[jj];
		sbuf[i].ranks[7 + kk] = g->rank;
	    }
	    else {
		sbuf[i].indexes[7 + kk] = g->owner_index_face[e->faces[jj]];
		sbuf[i].ranks[7 + kk] = g->owner_rank_face[e->faces[jj]];
	    }
	}
	/* the element types */
	sbuf[i].types[10] = g->types_elem[e->index] & ~OWNER;
	sbuf[i].indexes[10] = e->index;
	sbuf[i].ranks[10] = g->rank;
	/*----- copy element data */
	memcpy(&sbuf[i].elem, e, sizeof(*e));
	/*----- save peer element pointer in 'parent' */
	sbuf[i].elem.parent = g->neighbours.list[i].remote;
	/*----- save vertex + op_vertex in bits 0-5 of mark */
	sbuf[i].elem.mark = g->neighbours.list[i].vertex +
			    g->neighbours.list[i].op_vertex * 8;
	/*----- update with global indices */
	e = &sbuf[i].elem;
	for (ii = 0; ii < NVert; ii++)
	    e->verts[ii] = GlobalVertex(g, e->verts[ii]);
	for (ii = 0; ii < NEdge; ii++)
	    e->edges[ii] = GlobalEdge(g, e->edges[ii]);
	for (ii = 0; ii < NFace; ii++)
	    e->faces[ii] = GlobalFace(g, e->faces[ii]);
	e->index = GlobalElement(g, e->index);
    }

    /* exchange data */
    rbuf = phgAlloc(g->neighbours.count * sizeof(*rbuf));
    MPI_Type_contiguous(sizeof(*sbuf), MPI_BYTE, &type);
    MPI_Type_commit(&type);
    MPI_Alltoallv(sbuf, g->neighbours.counts, g->neighbours.displs, type,
		  rbuf, g->neighbours.counts, g->neighbours.displs, type,
		  g->comm);
    MPI_Type_free(&type);
    phgFree(sbuf);

    /* process received data */

    /* buffer for saving new vertices/edges/faces/elems.
     * note: at most 3 new items per element. */
    items = phgAlloc(g->neighbours.count * 3 * sizeof(*items));

    /* ordering buffer */
    n = g->nface;
    if (n < g->nedge)
	n = g->nedge;
    if (n < g->nvert)
	n = g->nvert;
    m = g->neighbours.count * 4;
    ordering = phgAlloc((n > m ? n : m) * sizeof(*ordering));

    /* ------------------- process incoming vertices ------------------- */
    n = g->nvert;
    L2Gmap = g->L2Gmap_vert;
    for (i = 0; i < n; i++)
	ordering[i] = i;
    qsort(ordering, n, sizeof(*ordering), comp_L2Gmap);
    m = 0;	/* number of (unsorted) new items */
    for (i = 0; i < g->neighbours.count; i++) {
	e = &rbuf[i].elem;
	vertex = rbuf[i].elem.mark % 8;	/* the face no of e */
	GetFaceVertices(e, vertex, u[0], u[1], u[2], ii);
	peer = rbuf[i].elem.parent;	/* peer element */
	peer_vertex = (rbuf[i].elem.mark / 8) % 8;
	GetFaceVertices(peer, peer_vertex, v[0], v[1], v[2], ii);
	/* update the 3 vertices on the common face */
	if (g->period == NULL ||
	    e->verts[u[0]] == GlobalVertex(g, peer->verts[v[0]])) {
	    /* interior face */
	    for (ii = 0; ii < 3; ii++) {
		assert(e->verts[u[ii]] == GlobalVertex(g, peer->verts[v[ii]]));
		e->verts[u[ii]] = peer->verts[v[ii]];
	    }
	}
	else {
	    /* face on the periodic boundary: get pindex from peer
	     * Note: the code relies on the fact that e->ordering was built
	     * using GlobalVertexP, such that matching vertices on the periodic
	     * face have the same global pindex (g->periodic->L2Gmap_vert) */
	    for (ii = 0; ii < 3; ii++) {
		k = e->verts[u[ii]];
		j = phgBinarySearch(n, L2Gmap, ordering, &k, sizeof(*ordering),
				    phgCompINT);
		if (j < n && k == L2Gmap[ordering[j]]
			  && g->types_vert[ordering[j]] != UNREFERENCED) {
		    /* found in L2Gmap[] */
		    e->verts[u[ii]] = ordering[j];
		    continue;
		}
		items[m].gindex = k;
		items[m].rbuf_index = i;
		items[m].item_index = u[ii];
		items[m].pindex = GlobalVertexP(g, peer->verts[v[ii]]);
		items[m].oindex = rbuf[i].indexes[u[ii]];
		items[m].orank = rbuf[i].ranks[u[ii]];
		items[m].type = rbuf[i].types[u[ii]];
		m++;
	    }
	}
	/* k = global no of the opposite vertex */
	k = e->verts[vertex];
	/* search for k in L2Gmap[] */
	j = phgBinarySearch(n, L2Gmap, ordering, &k, sizeof(*ordering),
			    phgCompINT);
	if (j < n && k == L2Gmap[ordering[j]]
		  && g->types_vert[ordering[j]] != UNREFERENCED) {
	    /* found in L2Gmap[] */
	    e->verts[vertex] = ordering[j];
	    continue;
	}
	items[m].gindex = k;
	items[m].rbuf_index = i;
	items[m].item_index = vertex;
	items[m].pindex = rbuf[i].pindex;
	items[m].oindex = rbuf[i].indexes[vertex];
	items[m].orank = rbuf[i].ranks[vertex];
	items[m].type = rbuf[i].types[vertex];
	m++;
    }
    if (m > 0) {
	/* sort list of items according to 'index' */
	for (i = 0; i < m; i++)
	    ordering[i] = i;
	/* sort list of items according to 'index' */
	qsort(ordering, m, sizeof(*ordering), comp_item);
	/* build halo->L2Gmap_vert[] and halo->verts[], and
	 * update g->{verts,L2Gmap_vert}, and e->verts[] entries in the halo */
	REALLOC(g->verts, g->nvert + m, g->nvert);
	REALLOC(g->L2Gmap_vert, g->nvert + m, g->nvert);
	if (g->period != NULL)
	    REALLOC(g->period->L2Gmap_vert, g->nvert + m, g->nvert);
	REALLOC(g->owner_index_vert, g->nvert + m, g->nvert);
	REALLOC(g->owner_rank_vert, g->nvert + m, g->nvert);
	REALLOC(g->types_vert, g->nvert + m, g->nvert);
	pi = items;
	n = g->nvert;
	for (i = 0; i < m; n++) {
	    pi = items + ordering[i];
	    memcpy(g->verts + n, rbuf[pi->rbuf_index].coord + pi->item_index,
		   sizeof(COORD));
	    j = g->L2Gmap_vert[n] = pi->gindex;
	    g->owner_index_vert[n] = pi->oindex;
	    g->owner_rank_vert[n] = pi->orank;
	    if (g->period != NULL) {
		g->period->L2Gmap_vert[n] = pi->pindex;
		if (g->owner_rank_vert[n] == g->rank) {
		    g->owner_index_vert[n] = -1;
		    g->owner_rank_vert[n] = -1;
		}
	    }
	    g->types_vert[n] = pi->type;
	    assert(g->owner_rank_vert[n] != g->rank);
	    do {
		pi = items + ordering[i];
		rbuf[pi->rbuf_index].elem.verts[pi->item_index] = n;
	    } while (++i < m && items[ordering[i]].gindex == j);
	}
	m += g->nvert;
	if (n < m) {
	    REALLOC(g->verts, n, m);
	    REALLOC(g->L2Gmap_vert, n, m);
	    if (g->period != NULL)
		REALLOC(g->period->L2Gmap_vert, n, m);
	    REALLOC(g->owner_index_vert, n, m);
	    REALLOC(g->owner_rank_vert, n, m);
	    REALLOC(g->types_vert, n, m);
	}
	phgInfo(1, "%s: nvert += %d\n", __func__, n - g->nvert);
	g->halo->nvert_bak = g->nvert;
	g->nvert = n;
	if (g->period != NULL) {
	    phgFree(g->period->ordering);
	    g->period->ordering = phgAlloc(n * sizeof(*ordering));
	    L2Gmap = g->period->L2Gmap_vert;
	    for (i = 0; i < n; i++)
		g->period->ordering[i] = i;
	    qsort(g->period->ordering, n, sizeof(*ordering), comp_L2Gmap);
	}
    }

    /* ------------------- process incoming edges ------------------- */
    n = g->nedge;
    L2Gmap = g->L2Gmap_edge;
    for (i = 0; i < n; i++)
	ordering[i] = i;
    qsort(ordering, n, sizeof(*ordering), comp_L2Gmap);
    m = 0;	/* number of (unsorted) new items */
    for (i = 0; i < g->neighbours.count; i++) {
	e = &rbuf[i].elem;
	vertex = rbuf[i].elem.mark % 8;	/* the face no of e */
	GetFaceVertices(e, vertex, u[0], u[1], u[2], ii);
	peer = rbuf[i].elem.parent;	/* peer element */
	peer_vertex = (rbuf[i].elem.mark / 8) % 8;
	GetFaceVertices(peer, peer_vertex, v[0], v[1], v[2], ii);
	/* update the 3 edges on the common face */
#define UpdateEdge(a,b) \
	ii = GetEdgeNo(u[a], u[b]); \
	jj = GetEdgeNo(v[a], v[b]); \
	assert(e->edges[ii] == GlobalEdge(g, peer->edges[jj])); \
	e->edges[ii] = peer->edges[jj];
	UpdateEdge(0,1);
	UpdateEdge(0,2);
	UpdateEdge(1,2);
	/* Search L2Gmap_edge[] for the three non common edges */
	for (ii = 0; ii < 3; ii++) {
	    jj = GetEdgeNo(vertex, u[ii]);
	    /* k = global no of the edge */
	    k = e->edges[jj];
	    /* search for k in L2Gmap[] */
	    j = phgBinarySearch(n, L2Gmap, ordering, &k, sizeof(*ordering),
				phgCompINT);
	    if (j < n && k == L2Gmap[ordering[j]]
		      && g->types_edge[ordering[j]] != UNREFERENCED) {
		/* found in L2Gmap[] */
		e->edges[jj] = ordering[j];
		continue;
	    }
	    items[m].gindex = k;
	    items[m].rbuf_index = i;
	    items[m].item_index = jj;
	    items[m].oindex = rbuf[i].indexes[4 + ii];
	    items[m].orank = rbuf[i].ranks[4 + ii];
	    items[m].type = rbuf[i].types[4 + ii];
	    m++;
	}
    }
    if (m > 0) {
	/* sort list of items according to 'index' */
	for (i = 0; i < m; i++)
	    ordering[i] = i;
	/* sort list of items according to 'index' */
	qsort(ordering, m, sizeof(*ordering), comp_item);
	/* build halo->L2Gmap_edge[], update e->edges[] entries in the halo */
	REALLOC(g->L2Gmap_edge, g->nedge + m, g->nedge);
	REALLOC(g->owner_index_edge, g->nedge + m, g->nedge);
	REALLOC(g->owner_rank_edge, g->nedge + m, g->nedge);
	REALLOC(g->types_edge, g->nedge + m, g->nedge);
	pi = items;
	n = g->nedge;
	for (i = 0; i < m; n++) {
	    pi = items + ordering[i];
	    j = g->L2Gmap_edge[n] = pi->gindex;
	    g->owner_index_edge[n] = pi->oindex;
	    g->owner_rank_edge[n] = pi->orank;
	    g->types_edge[n] = pi->type;
	    assert(g->owner_rank_edge[n] != g->rank);
	    do {
		pi = items + ordering[i];
		rbuf[pi->rbuf_index].elem.edges[pi->item_index] = n;
	    } while (++i < m && items[ordering[i]].gindex == j);
	}
	m += g->nedge;
	if (n < m) {
	    REALLOC(g->L2Gmap_edge, n, m);
	    REALLOC(g->owner_index_edge, n, m);
	    REALLOC(g->owner_rank_edge, n, m);
	    REALLOC(g->types_edge, n, m);
	}
	phgInfo(1, "%s: nedge += %d\n", __func__, n - g->nedge);
	g->halo->nedge_bak = g->nedge;
	g->nedge = n;
    }

    /* ------------------- process incoming faces ------------------- */
    n = g->nface;
    L2Gmap = g->L2Gmap_face;
    for (i = 0; i < n; i++)
	ordering[i] = i;
    qsort(ordering, n, sizeof(*ordering), comp_L2Gmap);
    m = 0;	/* number of (unsorted) new items */
    for (i = 0; i < g->neighbours.count; i++) {
	e = &rbuf[i].elem;
	vertex = rbuf[i].elem.mark % 8;	/* the face no of e */
	GetFaceVertices(e, vertex, u[0], u[1], u[2], ii);
	peer = rbuf[i].elem.parent;	/* peer element */
	peer_vertex = (rbuf[i].elem.mark / 8) % 8;
	GetFaceVertices(peer, peer_vertex, v[0], v[1], v[2], ii);
	/* update the the common face */
	assert(e->faces[vertex] == GlobalFace(g, peer->faces[peer_vertex]));
	e->faces[vertex] = peer->faces[peer_vertex];
	/* Search L2Gmap_face[] for the three non common faces */
	for (ii = 0; ii < 3; ii++) {
	    jj = u[ii];		/* face no */
	    /* k = global no of the face */
	    k = e->faces[jj];
	    /* search for k in L2Gmap[] */
	    j = phgBinarySearch(n, L2Gmap, ordering, &k, sizeof(*ordering),
				phgCompINT);
	    if (j < n && k == L2Gmap[ordering[j]]
		      && g->types_face[ordering[j]] != UNREFERENCED) {
		/* found in L2Gmap[] */
		e->faces[jj] = ordering[j];
		continue;
	    }
	    items[m].gindex = k;
	    items[m].rbuf_index = i;
	    items[m].item_index = jj;
	    items[m].oindex = rbuf[i].indexes[7 + ii];
	    items[m].orank = rbuf[i].ranks[7 + ii];
	    items[m].type = rbuf[i].types[7 + ii];
	    m++;
	}
    }
    if (m > 0) {
	/* sort list of items according to 'index' */
	for (i = 0; i < m; i++)
	    ordering[i] = i;
	/* sort list of items according to 'index' */
	qsort(ordering, m, sizeof(*ordering), comp_item);
	/* build halo->L2Gmap_face[], update e->faces[] entries in the halo */
	REALLOC(g->L2Gmap_face, g->nface + m, g->nface);
	REALLOC(g->owner_index_face, g->nface + m, g->nface);
	REALLOC(g->owner_rank_face, g->nface + m, g->nface);
	REALLOC(g->types_face, g->nface + m, g->nface);
	pi = items;
	n = g->nface;
	for (i = 0; i < m; n++) {
	    pi = items + ordering[i];
	    j = g->L2Gmap_face[n] = pi->gindex;
	    g->owner_index_face[n] = pi->oindex;
	    g->owner_rank_face[n] = pi->orank;
	    g->types_face[n] = pi->type;
	    assert(g->owner_rank_face[n] != g->rank);
	    do {
		pi = items + ordering[i];
		rbuf[pi->rbuf_index].elem.faces[pi->item_index] = n;
	    } while (++i < m && items[ordering[i]].gindex == j);
	}
	m += g->nface;
	if (n < m) {
	    REALLOC(g->L2Gmap_face, n, m);
	    REALLOC(g->owner_index_face, n, m);
	    REALLOC(g->owner_rank_face, n, m);
	    REALLOC(g->types_face, n, m);
	}
	phgInfo(1, "%s: nface += %d\n", __func__, n - g->nface);
	g->halo->nface_bak = g->nface;
	g->nface = n;
    }

    phgFree(items);

    /* ------------------- process incoming elements ------------------- */

    /* Note: since we don't look in g->L2Gmap_elem[], there may be multiple
     *	     entries in the enlarged list, i.e, there exists i such that:
     *		g->L2Gmap_elem[g->halo->nelem_bak + i] == g->L2Gmap_elem[i] */
    m = g->neighbours.count;
    if (m > 0) {
	/* sort list of items according to 'index' */
	for (i = 0; i < m; i++)
	    ordering[i] = i;
	/* sort list of items according to 'index' */
	qsort(ordering, m, sizeof(*ordering), comp_elem);
	/* build halo->L2Gmap_elem[], update e->index in the halo */
	g->halo->map = phgAlloc(m * sizeof(*g->halo->map));
	g->halo->elist = phgAlloc(m * sizeof(*g->halo->elist));
	REALLOC(g->L2Gmap_elem, g->nelem + m, g->nelem);
	REALLOC(g->owner_index_elem, g->nelem + m, g->nelem);
	REALLOC(g->owner_rank_elem, g->nelem + m, g->nelem);
	REALLOC(g->types_elem, g->nelem + m, g->nelem);
	REALLOC(g->elems, g->nelem + m, g->nelem);
	n = g->nelem;
	for (i = 0; i < m; n++) {
	    e = g->halo->elist + n - g->nelem;
	    k = ordering[i];
	    memcpy(e, &rbuf[k].elem, sizeof(ELEMENT));
	    e->index = n;
	    e->mark = 0;
	    e->parent = NULL;
	    j = g->L2Gmap_elem[n] = rbuf[k].elem.index;
	    g->owner_index_elem[n] = rbuf[k].indexes[10];
	    g->owner_rank_elem[n] = rbuf[k].ranks[10];
	    g->types_elem[n] = rbuf[k].types[10] & ~OWNER;
	    assert(g->owner_rank_elem[n] != g->rank);
	    g->elems[n] = e;
	    do {
		k = ordering[i];
		vertex = rbuf[k].elem.mark % 8;
		peer = rbuf[k].elem.parent;	/* peer element */
		peer_vertex = (rbuf[k].elem.mark / 8) % 8;
		/* update neighbour links */
		e->neighbours[vertex] = peer;
		assert(HasRemoteNeighbour(peer, peer_vertex));
		g->halo->map[(size_t)peer->neighbours[peer_vertex]]
							= n - g->nelem;
	    } while (++i < m && rbuf[ordering[i]].elem.index == j);
	}
	if (n < g->nelem + m) {
	    REALLOC(g->halo->elist, n - g->nelem, m);
	    m += g->nelem;
	    REALLOC(g->L2Gmap_elem, n, m);
	    REALLOC(g->owner_index_elem, n, m);
	    REALLOC(g->owner_rank_elem, n, m);
	    REALLOC(g->types_elem, n, m);
	    REALLOC(g->elems, n, m);
	    /* update pointers in g->elems[g->nelem:n] */
	    for (i = 0; i < n - g->nelem; i++) {
		e = g->halo->elist + i;
		assert(e->index >= g->nelem);
		g->elems[e->index] = e;
	    }
	}
	phgInfo(1, "%s: nelem += %d\n", __func__, n - g->nelem);
	g->halo->nelem_bak = g->nelem;
	g->nelem = n;
    }

    phgFree(ordering);
    phgFree(rbuf);

    if (g->nvert == g->halo->nvert_bak &&
	g->nedge == g->halo->nedge_bak &&
	g->nface == g->halo->nface_bak &&
	g->nelem == g->halo->nelem_bak)
	goto end;

    /* update HP_TYPEs */
    for (php = g->hp; php != NULL && *php != NULL; php++) {
	HP_TYPE *hp = *php;
	REALLOC(hp->elem_order, g->nelem, g->halo->nelem_bak);
	ForAllElements_(g, e, FOR_UNOWNED)
	    hp->elem_order[e->index] = hp->info->max_order;
	phgHPSetupOrders(hp);
    }

    /* adjust DOF data */
    for (pdof = g->dof; pdof!= NULL && *pdof != NULL; pdof++) {
	DOF *dof = *pdof;
	FLOAT *data, *p_old, *p_new;
	INT len_old;
	MAP *map;

	if (dof->data == NULL || SpecialDofType(dof->type))
	    continue;

	phgInfo(1, "%s: update DOF %s\n", __func__, dof->name);
	/* allocate new data buffer */
	p_new = data = phgCalloc(DofDataCount(dof), sizeof(*data));
	p_old = dof->data;

	len_old = DofVertexDataCount_(dof, g->halo->nvert_bak);
	memcpy(p_new, p_old, len_old * sizeof(*data));
	dof->data_vert = p_new;
	p_old += len_old;
	p_new += DofVertexDataCount(dof);

	len_old = DofEdgeDataCount_(dof, g->halo->nedge_bak);
	memcpy(p_new, p_old, len_old * sizeof(*data));
	dof->data_edge = p_new;
	p_old += len_old;
	p_new += DofEdgeDataCount(dof);

	len_old = DofFaceDataCount_(dof, g->halo->nface_bak);
	memcpy(p_new, p_old, len_old * sizeof(*data));
	dof->data_face = p_new;
	p_old += len_old;
	p_new += DofFaceDataCount(dof);

	len_old = DofElementDataCount_(dof, g->halo->nelem_bak);
	memcpy(p_new, p_old, len_old * sizeof(*data));
	dof->data_elem = p_new;

	phgFree(dof->data);
	dof->data = data;

	/* update halo data */

	if (dof == g->geom) {
	    /* Note: use FOR_UNOWNED to only compute halo data */
	    phgDofSetDataByFunction__(dof, NULL, NULL, NULL, ~0, FOR_UNOWNED);
	    continue;
	}

	if (update_dofs == FALSE || dof->userfunc == DofNoAction)
	    continue;

	if (dof->userfunc != DofInterpolation &&
	    (dof->userfunc != NULL || dof->userfunc_lambda != NULL)) {
	    phgDofSetDataByFunction__(dof,
		dof->userfunc, dof->userfunc_lambda, NULL, ~0, FOR_UNOWNED);
	    continue;
	}

	/* fetch data in the halo from remote processes */
	map = phgMapCreate(dof, NULL);
	data = phgAlloc(DofDataCount(dof) * sizeof(*data));
	phgMapDofToLocalData(map, 1, &dof, data);
	phgMapLocalDataToDof(map, 1, &dof, data);
	phgFree(data);
	phgMapDestroy(&map);
    }
end:
#else	/* USE_MPI */
    Unused(g);
    Unused(update_dofs);
#endif	/* USE_MPI */
    Unused(halo_type);

    /* rebuild face-to-element map */
    if (g->nface > 0 && g->Fmap_elems == NULL) {
	VEF_MAP *vef = phgDofSetupVEFMap_(g, NULL, FACE_FLAG, FOR_OWNED);
	phgDofFreeVEFMap(&vef);
    }
}

void
phgRemoveHalo(GRID *g)
{
    DOF **pdof;
    HP_TYPE **php;

    if (g->halo == NULL)
	return;

    if (g->nvert == g->halo->nvert_bak &&
	g->nedge == g->halo->nedge_bak &&
	g->nface == g->halo->nface_bak &&
	g->nelem == g->halo->nelem_bak)
	goto end;

    /* restore DOF data */
    for (pdof = g->dof; pdof!= NULL && *pdof != NULL; pdof++) {
	DOF *dof = *pdof;
	FLOAT *data, *p_old, *p_new;
	INT len, len_vert, len_edge, len_face, len_elem;
	if (dof->data == NULL || SpecialDofType(dof->type))
	    continue;

	phgInfo(1, "%s: update DOF %s\n", __func__, dof->name);

	/* new data sizes */
	len_vert = DofVertexDataCount_(dof, g->halo->nvert_bak);
	len_edge = DofEdgeDataCount_(dof, g->halo->nedge_bak);
	len_face = DofFaceDataCount_(dof, g->halo->nface_bak);
	len_elem = DofElementDataCount_(dof, g->halo->nelem_bak);

	/* allocate new data buffer */
	len = len_vert + len_edge + len_face + len_elem;
	p_new = data = phgAlloc(len * sizeof(*data));
	p_old = dof->data;

	memcpy(p_new, p_old, len_vert * sizeof(*data));
	dof->data_vert = p_new;
	p_old += DofVertexDataCount(dof);
	p_new += len_vert;

	memcpy(p_new, p_old, len_edge * sizeof(*data));
	dof->data_edge = p_new;
	p_old += DofEdgeDataCount(dof);
	p_new += len_edge;

	memcpy(p_new, p_old, len_face * sizeof(*data));
	dof->data_face = p_new;
	p_old += DofFaceDataCount(dof);
	p_new += len_face;

	memcpy(p_new, p_old, len_elem * sizeof(*data));
	dof->data_elem = p_new;

	phgFree(dof->data);
	dof->data = data;
    }

    /* restore HP_TYPEs */
    for (php = g->hp; php != NULL && *php != NULL; php++) {
	HP_TYPE *hp = *php;
	if ((hp->info->flags & VERT_FLAG)) {
	    REALLOC(hp->vert_order, g->halo->nvert_bak, g->nvert);
	    REALLOC(hp->vert_index, g->halo->nvert_bak + 1, g->nvert + 1);
	}
	if ((hp->info->flags & EDGE_FLAG)) {
	    REALLOC(hp->edge_order, g->halo->nedge_bak, g->nedge);
	    REALLOC(hp->edge_index, g->halo->nedge_bak + 1, g->nedge + 1);
	}
	if ((hp->info->flags & FACE_FLAG)) {
	    REALLOC(hp->face_order, g->halo->nface_bak, g->nface);
	    REALLOC(hp->face_index, g->halo->nface_bak + 1, g->nface + 1);
	}
	REALLOC(hp->elem_order, g->halo->nelem_bak, g->nelem);
	REALLOC(hp->elem_index, g->halo->nelem_bak + 1, g->nelem + 1);
    }

    /* invalidate face->element map */
    phgFree(g->Fmap_elems);
    g->Fmap_elems = NULL;
    phgFree(g->Fmap_faces);
    g->Fmap_faces = NULL;

    /* realloc buffers with new sizes */

#undef REALLOC
#define REALLOC(p) \
    p = phgRealloc_(p, g->halo->nvert_bak * sizeof(*p), g->nvert * sizeof(*p))
    if (g->nvert > g->halo->nvert_bak) {
	REALLOC(g->verts);
	REALLOC(g->types_vert);
#if USE_MPI
	REALLOC(g->L2Gmap_vert);
#endif	/* USE_MPI */
	REALLOC(g->owner_index_vert);
	REALLOC(g->owner_rank_vert);
	if (g->period != NULL) {
	    INT i, j;
	    REALLOC(g->period->L2Gmap_vert);
	    for (j = i = 0; i < g->nvert; i++) {
		if (g->period->ordering[i] >= g->halo->nvert_bak)
		    continue;
		g->period->ordering[j++] = g->period->ordering[i];
	    }
	    assert(j == g->halo->nvert_bak);
	    REALLOC(g->period->ordering);
	}
    }

#undef REALLOC
#define REALLOC(p) \
    p = phgRealloc_(p, g->halo->nedge_bak * sizeof(*p), g->nedge * sizeof(*p))
    if (g->nedge > g->halo->nedge_bak) {
	REALLOC(g->types_edge);
#if USE_MPI
	REALLOC(g->L2Gmap_edge);
	REALLOC(g->owner_index_edge);
	REALLOC(g->owner_rank_edge);
#endif	/* USE_MPI */
    }

#undef REALLOC
#define REALLOC(p) \
    p = phgRealloc_(p, g->halo->nface_bak * sizeof(*p), g->nface * sizeof(*p))
    if (g->nface > g->halo->nface_bak) {
	REALLOC(g->types_face);
#if USE_MPI
	REALLOC(g->L2Gmap_face);
#endif	/* USE_MPI */
    }

#undef REALLOC
#define REALLOC(p) \
    p = phgRealloc_(p, g->halo->nelem_bak * sizeof(*p), g->nelem * sizeof(*p))
    if (g->nelem > g->halo->nelem_bak) {
	REALLOC(g->elems);
	REALLOC(g->types_elem);
#if USE_MPI
	REALLOC(g->L2Gmap_elem);
#endif	/* USE_MPI */
    }

#if USE_MPI
    phgFree(g->owner_index_face);
    phgFree(g->owner_index_elem);
    g->owner_index_face = g->owner_index_elem = NULL;
    phgFree(g->owner_rank_face);
    phgFree(g->owner_rank_elem);
    g->owner_rank_face = g->owner_rank_elem = NULL;
#endif	/* USE_MPI */

    g->nvert = g->halo->nvert_bak;
    g->nedge = g->halo->nedge_bak;
    g->nface = g->halo->nface_bak;
    g->nelem = g->halo->nelem_bak;

end:
    phgFree(g->halo->elist);
    phgFree(g->halo->map);
    phgFree(g->halo->verts);
    phgFree(g->halo);
    g->halo = NULL;

    /* rebuild face-to-element map */
    if (g->nface > 0 && g->Fmap_elems == NULL) {
	VEF_MAP *vef = phgDofSetupVEFMap_(g, NULL, FACE_FLAG, FOR_OWNED);
	phgDofFreeVEFMap(&vef);
    }
}
